from documents.models import PageType, PageDirection, Page, ElementType, Element, ElementLink
from images.models import Image, ImageServer
from images.importer import bulk_zones, bulk_transcriptions
from abc import ABC, abstractmethod
from urllib.parse import urlsplit
import re
import roman
import logging
import os
import fnmatch
import uuid
import ijson


logger = logging.getLogger(__name__)

REGEX_FOLIO_PAGE = re.compile(r'^0*(\d+)(r|v)?')
REGEX_FOLIO_ROMAN = re.compile(r'^0*([MDCLXVI]+)(r|v)?')
FOLIO_WORDS = {
    'contre-plat': PageType.ContrePlat,
    'contre-plat inférieur': PageType.ContrePlat,
    'contre-plat supérieur': PageType.ContrePlat,
    'contre-garde': PageType.ContreGarde,
    'contre-garde supérieur': PageType.ContreGarde,
    'plat': PageType.Plat,
    'plat inférieur': PageType.Plat,
    'plat supérieur': PageType.Plat,
    'garde': PageType.Garde,
    'garde supérieur': PageType.Garde,
    'copie insérée': PageType.Copy,
}


def import_page(volume, image, register, folio, order):
    '''
    Build a page, inside a volume linked to a register and linked to an image via a zone
    '''
    assert isinstance(volume, Element) and volume.type == ElementType.Volume
    assert isinstance(image, Image)

    # Create a single register element holding all the pages
    element, _ = Element.objects.get_or_create(
        type=ElementType.Register,
        name=register,
    )
    ElementLink.objects.get_or_create(parent=volume, child=element)

    # Get the image zone if it exists
    poly = [[0, 0], [image.width, 0], [image.width, image.height], [0, image.height]]
    zone = image.zones.filter(
        polygon=poly,
        element__type=ElementType.Page
    )
    # Get the page or build zone and page
    if zone.exists():
        p = zone.first().element.page
    else:
        page_type, page_nb, page_direction = parse_folio(folio)
        p = Page.objects.create(
            folio=folio, name="Page {0} du volume {1}".format(folio, volume.name),
            page_type=page_type, nb=page_nb, direction=page_direction)
        p.zones.create(polygon=poly, image=image)

    ElementLink.objects.get_or_create(parent=element, child=p, defaults={'order': order})
    return p


def parse_folio(folio):
    """
    Parse folio to build a DB representation
    Outputs (type, nb, direction)
    """
    if folio is None:
        return (None, None, None)

    def _direction(d):
        if d is None:
            return
        if d.lower() in ('r', 'recto'):
            return PageDirection.Recto
        if d.lower() in ('v', 'verso'):
            return PageDirection.Verso

    # Standard page
    # Removing zeros, because this would match even roman numerals
    page_match = REGEX_FOLIO_PAGE.match(folio.lstrip('0'))
    if page_match is not None:
        groups = page_match.groups()
        page = int(groups[0])
        direction = _direction(groups[1])
        return (PageType.Page, page, direction)

    # Roman page
    roman_match = REGEX_FOLIO_ROMAN.match(folio)
    if roman_match is not None:
        groups = roman_match.groups()
        page = roman.fromRoman(groups[0])
        direction = _direction(groups[1])
        return (PageType.Summary, page, direction)

    # Words
    page_type = FOLIO_WORDS.get(folio.lower())

    return (page_type, None, None)


class ManifestsImporter(ABC):
    """
    Parses JSON manifests and annotation data to import them in the database.
    """
    offline = False

    def __init__(self, imgserv):
        """Initialize a manifest importer
        `imgserv` can be either one ImageServer or a list of ImageServers."""
        if isinstance(imgserv, ImageServer):
            self.imgserv = [imgserv]
        else:
            assert all(isinstance(srv, ImageServer) for srv in imgserv)
            self.imgserv = imgserv

        # This dictionary associates canvas IDs with images and pages
        # Filled by parse_manifest ; used by parse_annotation_list
        self.canvas_image = {}

        # This dictionary holds parsed data for each image's zones and transcriptions
        # Filled by parse_annotation_list ; used by save_transcriptions
        self.images_transcription_data = {}

        # Image server statistics - hold image counts for each server
        self.server_images_count = {}

    @abstractmethod
    def get_json_files(self):
        """
        Generator yielding a stream (any class that has read() and seek(int) methods)
        for each JSON file to parse
        """
        pass

    def run(self, offline=False):
        """Start the importing process."""
        self.offline = offline
        for f in self.get_json_files():
            self.parse_file(f)
            self.save_transcriptions()
        for srv, count in self.server_images_count.items():
            logger.info("Saved {0} images for server {1} ({2})".format(
                count, srv.name, srv.url))

    def parse_file(self, stream):
        """Parse a JSON file loaded in a stream."""
        parser = ijson.parse(stream)
        for prefix, event, value in parser:
            if (prefix, event) == ('@type', 'string'):
                if value == 'sc:Manifest':
                    stream.seek(0)
                    self.parse_manifest(stream)
                    break
                elif value == 'sc:AnnotationList':
                    stream.seek(0)
                    self.parse_annotation_list(stream)
                    break

    def parse_manifest(self, stream):
        """Parse a IIIF manifest loaded as a stream."""
        # Get this file's volume range ID from the top-most structure
        range_id = next(struct['ranges'][0]
                        for struct in ijson.items(stream, "structures.item")
                        if struct.get('viewingHint') == "top")
        stream.seek(0)

        # Get our volume's structure and label
        vol_struct = next(struct for struct in ijson.items(stream, "structures.item")
                          if struct.get('@id') == range_id)
        vol_name = vol_struct['label']
        logger.debug("Creating volume {}".format(vol_name))

        # Create a volume and a register
        vol, _ = Element.objects.get_or_create(name=vol_name, type=ElementType.Volume)
        doc, _ = Element.objects.get_or_create(name=vol_name, type=ElementType.Register)
        ElementLink.objects.get_or_create(parent=vol, child=doc)
        stream.seek(0)
        self.parse_manifest_canvases(stream, vol)

    def parse_manifest_canvases(self, stream, volume):
        """Parse canvases in a IIIF manifest stream and
        create images and pages with a given volume name."""
        # Enumerating to use index as order number in ElementLink
        for i, canvas in enumerate(ijson.items(stream, 'sequences.item.canvases.item')):
            # Label contains the folio
            if canvas.get('label') is None:
                logger.warning(
                    "Found an image canvas with id {0} for volume {1}, "
                    "but no folio (label) was specified".format(
                        canvas.get('@id'), str(volume)))
                continue
            folio = canvas['label']

            # Get the image resource
            resource = canvas['images'][0].get('resource')
            if resource is None:
                logger.warning(
                    "Found an image canvas with id {0} for volume {1}, "
                    "but no image resource was specified".format(
                        canvas.get('@id'), str(volume)))
                continue

            # Go find the service ID to get the image URL
            if resource.get('service') is None:
                logger.warning(
                    "Found an image resource with id {}, but no service was specified".format(
                        resource.get('@id')))
                continue
            service_id = resource['service'].get('@id')

            # Find the right server
            image_server = self.find_image_server(service_id)
            assert image_server is not None, "No image server found for image {}".format(service_id)
            # Strip server URL to get image path
            path = service_id[len(image_server.url):].lstrip('/')

            # Create image and page
            try:
                logger.debug("Looking for image {}".format(path))
                image = image_server.find_image(path, offline=self.offline)
                logger.debug("Found image {}".format(image.id))
                logger.debug("Looking for page {}".format(folio))
                page = import_page(volume, image, volume.name, folio, i)
                logger.debug("Imported page {}".format(page.id))
            except Exception as e:
                logger.error('Failed to import page: {}'.format(e))
                return

            # Fill canvas_image dictionary for usage by annotation list parsing
            self.canvas_image[canvas.get('@id')] = (image, page)

            # Fill image server statistics
            self.server_images_count[image_server] = self.server_images_count.get(image_server, 0) + 1

            logger.info('Added page {} - image {}'.format(page, image))

    def parse_annotation_list(self, stream):
        """Parse a IIIF annotation list loaded as a stream.
        Zones and transcriptions are not immediately saved; use save_transcriptions()."""
        for annotation in ijson.items(stream, 'resources.item'):
            # Check annotation contains text
            if annotation['resource'].get('chars') is None or \
                    annotation['resource']['@type'] != 'cnt:ContentAsText' or \
                    annotation['resource']['format'] != 'text/plain':
                logger.warning("Incompatible annotation with ID {}".format(
                    annotation.get('@id')))
                continue
            text = annotation['resource']['chars']

            # Get ID and zone information from URI
            canvas_id, data = annotation['on'].split('#', 1)

            # Find the associated image
            if canvas_id not in self.canvas_image:
                logger.warning(
                    "Could not find an associated image for annotation with ID {}".format(
                        annotation.get('@id')))
                continue
            image, page = self.canvas_image[canvas_id]
            # Add to dictionary
            if self.images_transcription_data.get((image, page)) is None:
                self.images_transcription_data[(image, page)] = []

            # Parse the data string (format is xywh=0,100,200,300) and cast to int
            # rpartition("=")[2] gives the part after the '=' and is faster than a split
            x, y, w, h = [int(v) for v in data.rpartition('=')[2].split(',')]

            # Fill data dictionary
            self.images_transcription_data[(image, page)].append({
                'x': x, 'y': y, 'width': w, 'height': h, 'text': text, 'line': None, 'score': 1
            })

    def save_transcriptions(self):
        """To optimize transcription parsing, saving and indexing is done in bulk."""
        total_zones, total_transcriptions, total_indexes = 0, 0, 0

        for (image, page), data in self.images_transcription_data.items():
            new_zones = bulk_zones(image, data)
            total_zones += len(new_zones)
            logger.debug("Created {0} zones for image {1}".format(len(new_zones), image.path))

            new_transcriptions = bulk_transcriptions(image, page, data)
            total_transcriptions += len(new_transcriptions)
            logger.debug("Created {0} transcriptions for image {1}".format(len(new_transcriptions), image.path))

            index_count = image.index()
            total_indexes += index_count
            logger.debug("Created {0} indexes for image {1}".format(index_count, image.path))

        logger.info("Saved {0} zones, {1} transcriptions and {2} indexes".format(
            total_zones, total_transcriptions, total_indexes))

        # Reset
        self.images_transcription_data = {}

    def find_image_server(self, image_url):
        """Find an image server for a specific image.
        If server is not found, will ask the user; if user refuses to create a server, will return None."""
        try:  # Look in the manually specified servers
            return next(srv for srv in self.imgserv if image_url.startswith(srv.url))
        except StopIteration:
            pass

        try:  # Look in the database
            return next(srv for srv in ImageServer.objects.all() if image_url.startswith(srv.url))
        except StopIteration:
            pass

        # See if the server path can be automatically detected
        splat = urlsplit(image_url)
        if splat.path.startswith('/iiif'):
            new_url = "{0}://{1}/iiif".format(splat.scheme, splat.netloc)
            new_name = str(uuid.uuid4())[:5]
            new_srv = ImageServer.objects.create(name=new_name, url=new_url)
            logger.info(
                "Created IIIF image server '{0}' with URL '{1}'".format(new_name, new_url))
            return new_srv

        # Ask for help
        print("Unknown server for image with URL '{}'".format(image_url))
        if self._boolean_input("Create it in database? [y/N]", default=False):
            new_server_url, new_server_name = "", ""

            while len(new_server_url) < 1 or not image_url.startswith(new_server_url):
                new_server_url = input("Server URL : ")
                if not image_url.startswith(new_server_url):
                    print("This server URL does not match the image's URL.")

            while len(new_server_name) < 1:
                new_server_name = input("Server name : ")

            return ImageServer.objects.create(name=new_server_name, url=new_server_url)
        return None

    def _boolean_input(self, question, default=None):
        """Ask something and wait for Yes or No
        Copied from Django's source code"""
        result = input("%s " % question)
        if not result and default is not None:
            return default
        while len(result) < 1 or result[0].lower() not in "yn":
            result = input("Please answer yes or no: ")
        return result[0].lower() == "y"


class LocalManifestsImporter(ManifestsImporter):
    """Allows importing of local JSON files."""

    def __init__(self, imgserv, path):
        super().__init__(imgserv)
        self.path = path

    def get_json_files(self):
        # Support single file & directories
        if os.path.isdir(self.path):
            paths = [
                os.path.join(root, filename)
                for root, _, filenames in os.walk(self.path)
                for filename in fnmatch.filter(filenames, "*.json")
            ]
        else:
            paths = [os.path.realpath(self.path), ]

        for path in paths:
            logger.info("Opening {}".format(path))
            yield open(path, 'rb')
