from rest_framework import serializers
from django.conf import settings
from documents.models import Element, Transcription, Page, ElementType
from images.models import Image, Zone
from images.serializers import ZoneSerializer, ImageSerializer
from arkindex.tools import sslify_url


class EnumField(serializers.Field):
    """
    Serializes an enum field into a JSON value
    """

    def to_representation(self, obj):
        return obj.value

    def to_internal_value(self, data):
        # TODO !
        return


class TranscriptionSerializer(serializers.ModelSerializer):
    """
    Serialises a Transcription
    issued from a search
    """
    class Meta:
        model = Transcription
        fields = (
            'id',
            'text',
            'line',
            'score',
        )


class ElementLightSerializer(serializers.ModelSerializer):
    """
    Serialises a Element
    """
    type = EnumField()

    class Meta:
        model = Element
        fields = (
            'id',
            'type',
            'name',
        )


class PageLightSerializer(serializers.ModelSerializer):
    """
    Serialises a Page
    """
    page_type = EnumField()
    direction = EnumField()
    images = ImageSerializer(many=True)

    class Meta:
        model = Page
        fields = (
            'id',
            'page_type',
            'nb',
            'direction',
            'display_name',
            'images',
        )


class ElementSerializer(serializers.ModelSerializer):
    """
    Fully Serialises a document
    """
    type = EnumField()

    class Meta:
        model = Element
        fields = (
            'id',
            'type',
            'name',
        )


class SearchResultSerializer(serializers.ModelSerializer):
    """
    Link between objects & their search indexation
    """
    zones = ZoneSerializer(many=True)
    parents = ElementLightSerializer(source='parent_docs', many=True)

    class Meta:
        model = Transcription
        fields = (
            'id',
            'text',
            'line',
            'score',
            'zones',
            'parents',
        )


class VolumeManifestSerializer(serializers.BaseSerializer):
    """
    Serialize a volume into a IIIF manifest
    """

    def to_representation(self, volume):
        assert isinstance(volume, Element) and volume.type == ElementType.Volume
        assert 'request' in self.context, "A request is required to generate absolute URLs"
        zones = []
        for p in Page.objects.get_descending(volume.id):
            zones.extend(list(p.zones.all()))
        canvases = PageZoneCanvasManifestSerializer(
            zones,
            context=self.context,
            many=True
        ).data
        return {
            "@context": settings.IIIF_PRESENTATION_CONTEXT,
            "@id": volume.build_absolute_url(self.context['request'], 'api:volume-manifest'),
            "@type": "sc:Manifest",
            "thumbnail": ImageThumbnailManifestSerializer(volume.get_thumbnail()).data,
            "related": [],
            "structures": self._get_structures(volume, canvases),
            "description": "",
            "sequences": [
                {
                    "canvases": canvases,
                    "label": "Reproduction intégrale",
                    "@id": volume.build_absolute_url(self.context['request'], 'api:sequence-manifest'),
                    "@type": "sc:Sequence"
                }
            ],
            "viewingHint": "individuals",
            "label": volume.name,
            "viewingDirection": "left-to-right"
        }

    def _get_structures(self, volume, canvases):
        return [{
            "viewingHint": "top",
            "label": volume.name,
            "@id": volume.build_absolute_url(self.context['request'], 'api:volume-manifest'),
            "ranges": [c['@id'] for c in canvases],
            "@type": "sc:Range"
        }] + [
            {
                "canvases": [c['@id']],
                "label": c['label'],
                "@id": c['@id'],
                "within": c['@id'],
                "@type": "sc:Range"
            }
            for c in canvases]


class ImageThumbnailManifestSerializer(serializers.BaseSerializer):
    """
    Serialize an image into a IIIF manifest thumbnail
    """

    def to_representation(self, image):
        assert isinstance(image, Image)
        return {
            "@id": sslify_url(image.get_thumbnail_url()),
            "service": {
                "@context": settings.IIIF_IMAGE_CONTEXT,
                "@id": sslify_url(image.url),
                "profile": settings.IIIF_SERVICE_PROFILE
            }
        }


class PageZoneCanvasManifestSerializer(serializers.BaseSerializer):
    """
    Serialize a page zone into a IIIF canvas
    """

    def to_representation(self, zone):
        assert isinstance(zone, Zone)
        assert isinstance(zone.element, Page)
        page = zone.element
        assert 'request' in self.context, "A request is required to generate absolute URLs"
        return {
            "@id": page.build_absolute_url(self.context['request'], 'api:canvas-manifest'),
            "@type": "sc:Canvas",
            "label": page.display_name,
            "height": zone.box.height,
            "width": zone.box.width,
            "images": [
                {
                    "@type": "oa:Annotation",
                    "resource": ImageResourceManifestSerializer(zone.image, context=self.context).data,
                    "on": page.build_absolute_url(self.context['request'], 'api:canvas-manifest'),
                    "motivation": "sc:painting"
                }
            ],
            "otherContent": [
                {
                    "@id": page.build_absolute_url(self.context['request'], 'api:page-manifest'),
                    "@type": "sc:AnnotationList",
                    "label": "Transcriptions"
                }
            ]
        }


class ImageResourceManifestSerializer(serializers.BaseSerializer):
    """
    Serialize an image into a IIIF resource.
    """

    def to_representation(self, image):
        assert isinstance(image, Image)
        return {
            "@id": sslify_url(image.get_thumbnail_url(max_width=None, max_height=None)),
            "@type": "dctypes:Image",
            "height": image.height,
            "width": image.width,
            "service": {
                "@context": settings.IIIF_IMAGE_CONTEXT,
                "@id": sslify_url(image.url),
                "profile": settings.IIIF_SERVICE_PROFILE
            }
        }


class PageAnnotationListSerializer(serializers.BaseSerializer):
    """
    Serialize an page's transcriptions into a IIIF annotation list
    """
    def to_representation(self, page):
        assert isinstance(page, Page)
        assert 'request' in self.context, "A request is required to generate absolute URLs"

        transcriptions = []
        for zone in page.zones.all():
            transcriptions.extend(Transcription.objects.filter(zones__in=zone.image.zones.all()))

        return {
            "@context": settings.IIIF_PRESENTATION_CONTEXT,
            "@id": self.context['request'].build_absolute_uri(),
            "@type": "sc:AnnotationList",
            "resources": TranscriptionAnnotationSerializer(
                transcriptions, context=self.context, many=True).data
        }


class TranscriptionAnnotationSerializer(serializers.BaseSerializer):
    """
    Serialize a transcription into a IIIF annotation
    """

    def to_representation(self, ts):
        assert isinstance(ts, Transcription)
        assert 'request' in self.context, "A request is required to generate absolute URLs"
        tid = ts.build_absolute_url(
            self.context['request'], 'api:transcription-manifest',
            id_argument='page_pk', transcription_pk=ts.id)
        zone = ts.zones.first()
        return {
            "@id": tid,
            "@type": "oa:Annotation",
            "motivation": "sc:painting",
            "on": zone.image.get_thumbnail_url() + "#xywh={0.x},{0.y},{0.width},{0.height}".format(zone.box),
            "resource": {
                "@id": tid,
                "@type": "cnt:ContentAsText",
                "chars": ts.text,
                "format": "text/plain"
            }
        }
