from images.models import Zone, Image
from documents.models import Transcription, ElementLink
from arkindex.tools import BoundingBox
import os
import re
import gzip
import logging

REGEX_INDEX = re.compile(
    b'^(?:line_(\d+) )?(.+) \d+ ([\de\-\.]+) (\d+) (\d+) (\d+) (\d+)')

logger = logging.getLogger(__name__)


def import_indexes(image, index_path, extension='jpg'):
    """
    Import indexes from file
    One gzipped index file per image
    Format: #line word n/a score-[0,1] x1 y1 x2 y2
    """
    assert isinstance(image, Image)
    assert os.path.exists(index_path), \
        'Missing index {}'.format(index_path)
    assert index_path.endswith('.idx.gz')

    # Parse lines content
    lines = []
    with gzip.open(index_path) as f:
        for line in f.readlines():
            if line.decode('utf-8').startswith('#'):
                continue
            index = REGEX_INDEX.match(line)
            if index is None:
                logger.warn('Index parsing failed : {}'.format(line))
                continue

            # Build zone
            data = index.groups()
            x, y = int(data[3]), int(data[4])
            lines.append({
                'x': x,
                'y': y,
                'width': int(data[5]) - x,
                'height': int(data[6]) - y,
                'text': data[1].decode('utf-8'),
                'line': int(data[0] or 0),
                'score': float(data[2]),
            })

    logger.info('Parsed {} lines'.format(len(lines)))

    # Create zones
    new_zones = bulk_zones(image, lines)
    logger.info('Created {} zones'.format(len(new_zones)))

    # Create transcriptions
    new_transcriptions = bulk_transcriptions(image, lines)
    logger.info('Created {} transcriptions '.format(len(new_transcriptions)))

    # Index all transcriptions into ES
    nb_inserts = image.index()
    logger.info('Added {} ES indexes'.format(nb_inserts))


def bulk_zones(image, items):
    """
    Create zones in bulk (one SQL statement)
    This is WAY faster
    """
    # Load existing zones in images
    existing = [
        BoundingBox(p)
        for p in image.zones.all().values_list('polygon', flat=True)
    ]
    existing = [
        (b.x, b.y, b.x + b.width, b.y + b.height) for b in existing
    ]

    # Calc needed insert
    needed = set([
        (item['x'], item['y'], item['x'] + item['width'], item['y'] + item['height'])
        for item in items
    ]).difference(existing)

    # Bulk create in db
    return Zone.objects.bulk_create([
        Zone(image=image, polygon=[[z[0], z[1]], [z[2], z[1]], [z[2], z[3]], [z[0], z[3]]])
        for z in needed
    ])


def bulk_transcriptions(image, page, items):
    """
    Create transcriptions in bulk
    """

    # Index existing zones, by unique keys
    zones = {
        repr(BoundingBox(z[1])): z[0]
        for z in image.zones.all().values_list('id', 'polygon')
    }

    # Load existing transcriptions
    existing = Transcription.objects.filter(
        zones__image=image
    ).values_list('zones__id', 'line', 'text', 'score')

    # Calc needed insert
    needed = set([
        (
            zones[repr(BoundingBox(
                [[i['x'], i['y']],
                 [i['x'] + i['width'], i['y'] + i['height']]]
            ))],
            i['line'],
            i['text'],
            i['score'],
        )
        for i in items
    ]).difference(existing)

    # Create transcriptions and associate zones
    all_ts = []
    for t in needed:
        ts = Transcription.objects.create(line=t[1], text=t[2], score=t[3])
        all_ts.append(ts)
        z = image.zones.get(id=t[0])
        z.element_id = ts.id
        z.save()
    max_order_dl = ElementLink.objects.filter(parent=page).order_by('-order').first()
    max_order = 0 if max_order_dl is None else max_order_dl.order + 1
    ElementLink.objects.bulk_create(
        ElementLink(parent=page, child=ts, order=i)
        for i, ts in enumerate(all_ts, max_order)
    )

    return all_ts
